@@ -27,12 +27,17 @@
 #define _GNU_SOURCE
 #include <linux/if.h>
 #include <linux/icmpv6.h>
+#include <linux/neighbour.h>
+#include <linux/if_ether.h>
+#include <net/if_arp.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
+#include <errno.h>
 
-#include "util.h"
+#include "useful.h"
 #include "netlink.h"
 #include "log.h"
+#include "util.h"
 #include "rtnl.h"
 #include "private.h"
 
@@ -145,6 +150,12 @@ LIB_EXPORT struct l_rtnl_address *l_rtnl_address_new(const char *ip,
 	return addr;
 }
 
+LIB_EXPORT struct l_rtnl_address *l_rtnl_address_clone(
+					const struct l_rtnl_address *orig)
+{
+	return l_memdup(orig, sizeof(struct l_rtnl_address));
+}
+
 LIB_EXPORT void l_rtnl_address_free(struct l_rtnl_address *addr)
 {
 	l_free(addr);
@@ -382,6 +393,16 @@ LIB_EXPORT uint8_t l_rtnl_route_get_family(const struct l_rtnl_route *rt)
 	return rt->family;
 }
 
+LIB_EXPORT bool l_rtnl_route_get_gateway(const struct l_rtnl_route *rt,
+						char *out_buf)
+{
+	if (unlikely(!rt))
+		return false;
+
+	return !address_to_string(rt->family, &rt->gw.in_addr, &rt->gw.in6_addr,
+					out_buf);
+}
+
 LIB_EXPORT uint32_t l_rtnl_route_get_lifetime(const struct l_rtnl_route *rt)
 {
 	if (unlikely(!rt))
@@ -507,7 +528,7 @@ LIB_EXPORT bool l_rtnl_route_set_protocol(struct l_rtnl_route *rt,
 	return true;
 }
 
-uint8_t l_rtnl_route_get_scope(const struct l_rtnl_route *rt)
+LIB_EXPORT uint8_t l_rtnl_route_get_scope(const struct l_rtnl_route *rt)
 {
 	if (unlikely(!rt))
 		return RT_SCOPE_NOWHERE;
@@ -515,7 +536,7 @@ uint8_t l_rtnl_route_get_scope(const struct l_rtnl_route *rt)
 	return rt->scope;
 }
 
-bool l_rtnl_route_set_scope(struct l_rtnl_route *rt, uint8_t scope)
+LIB_EXPORT bool l_rtnl_route_set_scope(struct l_rtnl_route *rt, uint8_t scope)
 {
 	if (unlikely(!rt))
 		return false;
@@ -745,6 +766,7 @@ LIB_EXPORT uint32_t l_rtnl_set_powered(struct l_netlink *rtnl, int ifindex,
 LIB_EXPORT void l_rtnl_ifaddr4_extract(const struct ifaddrmsg *ifa, int bytes,
 				char **label, char **ip, char **broadcast)
 {
+	char buf[INET_ADDRSTRLEN];
 	struct in_addr in_addr;
 	struct rtattr *attr;
 
@@ -756,7 +778,8 @@ LIB_EXPORT void l_rtnl_ifaddr4_extract(const struct ifaddrmsg *ifa, int bytes,
 				break;
 
 			in_addr = *((struct in_addr *) RTA_DATA(attr));
-			*ip = l_strdup(inet_ntoa(in_addr));
+			*ip = l_strdup(inet_ntop(AF_INET, &in_addr, buf,
+							INET_ADDRSTRLEN));
 
 			break;
 		case IFA_BROADCAST:
@@ -764,7 +787,8 @@ LIB_EXPORT void l_rtnl_ifaddr4_extract(const struct ifaddrmsg *ifa, int bytes,
 				break;
 
 			in_addr = *((struct in_addr *) RTA_DATA(attr));
-			*broadcast = l_strdup(inet_ntoa(in_addr));
+			*broadcast = l_strdup(inet_ntop(AF_INET, &in_addr, buf,
+							INET_ADDRSTRLEN));
 
 			break;
 		case IFA_LABEL:
@@ -1125,6 +1149,59 @@ done:
 	return id;
 }
 
+LIB_EXPORT struct l_rtnl_address *l_rtnl_ifaddr_extract(
+						const struct ifaddrmsg *ifa,
+						int bytes)
+{
+	struct rtattr *attr;
+	struct ifa_cacheinfo *cinfo;
+	struct l_rtnl_address *addr;
+
+	if (unlikely(!ifa))
+		return NULL;
+
+	if (!L_IN_SET(ifa->ifa_family, AF_INET, AF_INET6))
+		return NULL;
+
+	addr = l_new(struct l_rtnl_address, 1);
+	addr->prefix_len = ifa->ifa_prefixlen;
+	addr->family = ifa->ifa_family;
+	addr->flags = ifa->ifa_flags;
+	addr->scope = ifa->ifa_scope;
+
+	for (attr = IFA_RTA(ifa); RTA_OK(attr, bytes);
+						attr = RTA_NEXT(attr, bytes)) {
+		switch (attr->rta_type) {
+		case IFA_LOCAL:
+			if (ifa->ifa_family == AF_INET)
+				addr->in_addr =
+					*((struct in_addr *) RTA_DATA(attr));
+
+			break;
+		case IFA_ADDRESS:
+			if (ifa->ifa_family == AF_INET6)
+				addr->in6_addr =
+					*((struct in6_addr *) RTA_DATA(attr));
+
+			break;
+		case IFA_BROADCAST:
+			addr->broadcast = *((struct in_addr *) RTA_DATA(attr));
+			break;
+		case IFA_LABEL:
+			l_strlcpy(addr->label, RTA_DATA(attr),
+					sizeof(addr->label));
+			break;
+		case IFA_CACHEINFO:
+			cinfo = RTA_DATA(attr);
+			addr->preferred_lifetime = cinfo->ifa_prefered;
+			addr->valid_lifetime = cinfo->ifa_valid;
+			break;
+		}
+	}
+
+	return addr;
+}
+
 LIB_EXPORT uint32_t l_rtnl_ifaddr_add(struct l_netlink *rtnl, int ifindex,
 					const struct l_rtnl_address *addr,
 					l_netlink_command_func_t cb,
@@ -1238,3 +1315,126 @@ LIB_EXPORT uint32_t l_rtnl_route_delete(struct l_netlink *rtnl, int ifindex,
 	return _rtnl_route_change(rtnl, RTM_DELROUTE, ifindex, rt,
 						cb, user_data, destroy);
 }
+
+struct rtnl_neighbor_get_data {
+	l_rtnl_neighbor_get_cb_t cb;
+	void *user_data;
+	l_netlink_destroy_func_t destroy;
+};
+
+static void rtnl_neighbor_get_cb(int error, uint16_t type, const void *data,
+					uint32_t len, void *user_data)
+{
+	struct rtnl_neighbor_get_data *cb_data = user_data;
+	const struct ndmsg *ndmsg = data;
+	struct rtattr *attr;
+	const uint8_t *hwaddr = NULL;
+	size_t hwaddr_len = 0;
+
+	if (error != 0)
+		goto done;
+
+	if (type != RTM_NEWNEIGH || len < NLMSG_ALIGN(sizeof(*ndmsg))) {
+		error = -EIO;
+		goto done;
+	}
+
+	if (!(ndmsg->ndm_state & (NUD_PERMANENT | NUD_NOARP | NUD_REACHABLE))) {
+		error = -ENOENT;
+		goto done;
+	}
+
+	attr = (void *) ndmsg + NLMSG_ALIGN(sizeof(*ndmsg));
+	len -= NLMSG_ALIGN(sizeof(*ndmsg));
+
+	for (; RTA_OK(attr, len); attr = RTA_NEXT(attr, len))
+		switch (attr->rta_type) {
+		case NDA_LLADDR:
+			hwaddr = RTA_DATA(attr);
+			hwaddr_len = RTA_PAYLOAD(attr);
+			break;
+		}
+
+	if (!hwaddr)
+		error = -EIO;
+
+done:
+	if (cb_data->cb) {
+		cb_data->cb(error, hwaddr, hwaddr_len, cb_data->user_data);
+		cb_data->cb = NULL;
+	}
+}
+
+static void rtnl_neighbor_get_destroy_cb(void *user_data)
+{
+	struct rtnl_neighbor_get_data *cb_data = user_data;
+
+	if (cb_data->destroy)
+		cb_data->destroy(cb_data->user_data);
+
+	l_free(cb_data);
+}
+
+LIB_EXPORT uint32_t l_rtnl_neighbor_get_hwaddr(struct l_netlink *rtnl,
+					int ifindex, int family,
+					const void *ip,
+					l_rtnl_neighbor_get_cb_t cb,
+					void *user_data,
+					l_netlink_destroy_func_t destroy)
+{
+	size_t bufsize = NLMSG_ALIGN(sizeof(struct ndmsg)) +
+			RTA_SPACE(16); /* NDA_DST */
+	uint8_t buf[bufsize];
+	struct ndmsg *ndmsg = (struct ndmsg *) buf;
+	void *rta_buf = (void *) ndmsg + NLMSG_ALIGN(sizeof(struct ndmsg));
+	__auto_type cb_data = struct_alloc(rtnl_neighbor_get_data,
+						cb, user_data, destroy);
+	uint32_t ret;
+
+	memset(buf, 0, bufsize);
+	ndmsg->ndm_family = family;
+	ndmsg->ndm_ifindex = ifindex;
+	ndmsg->ndm_flags = 0;
+
+	rta_buf += rta_add_address(rta_buf, NDA_DST, family, ip, ip);
+
+	ret = l_netlink_send(rtnl, RTM_GETNEIGH, 0, ndmsg,
+				rta_buf - (void *) ndmsg,
+				rtnl_neighbor_get_cb, cb_data,
+				rtnl_neighbor_get_destroy_cb);
+	if (ret)
+		return ret;
+
+	l_free(cb_data);
+	return 0;
+}
+
+LIB_EXPORT uint32_t l_rtnl_neighbor_set_hwaddr(struct l_netlink *rtnl,
+					int ifindex, int family,
+					const void *ip,
+					const uint8_t *hwaddr,
+					size_t hwaddr_len,
+					l_netlink_command_func_t cb,
+					void *user_data,
+					l_netlink_destroy_func_t destroy)
+{
+	size_t bufsize = NLMSG_ALIGN(sizeof(struct ndmsg)) +
+			RTA_SPACE(16) +        /* NDA_DST */
+			RTA_SPACE(hwaddr_len); /* NDA_LLADDR */
+	uint8_t buf[bufsize];
+	struct ndmsg *ndmsg = (struct ndmsg *) buf;
+	void *rta_buf = (void *) ndmsg + NLMSG_ALIGN(sizeof(struct ndmsg));
+
+	memset(buf, 0, bufsize);
+	ndmsg->ndm_family = family;
+	ndmsg->ndm_ifindex = ifindex;
+	ndmsg->ndm_flags = 0;
+	ndmsg->ndm_state = NUD_REACHABLE;
+
+	rta_buf += rta_add_address(rta_buf, NDA_DST, family, ip, ip);
+	rta_buf += rta_add_data(rta_buf, NDA_LLADDR, hwaddr, hwaddr_len);
+
+	return l_netlink_send(rtnl, RTM_NEWNEIGH, NLM_F_CREATE | NLM_F_REPLACE,
+				ndmsg, rta_buf - (void *) ndmsg,
+				cb, user_data, destroy);
+}
